import type { PathCommand } from '@antv/g';
import { Rect } from '@antv/g';
import { clone, deepMix, isNumber, isArray, isFunction } from '@antv/util';
import { Linear, Band } from '@antv/scale';
import { GUI } from '../core/gui';
import { Lines } from './lines';
import { Columns } from './columns';
import { getRange, getStackedData } from './utils';
import type { ILinesCfg } from './lines';
import type { IColumnCfg, IColumnsCfg } from './columns';
import type { Data, SparklineCfg, SparklineOptions } from './types';
import {
  dataToLines,
  lineToLinePath,
  lineToCurvePath,
  linesToAreaPaths,
  linesToStackAreaPaths,
  linesToStackCurveAreaPaths,
} from './path';

export type { SparklineCfg, SparklineOptions };

export class Sparkline extends GUI<SparklineCfg> {
  public static tag = 'Sparkline';

  // sparkline容器
  private containerShape!: Rect;

  // Lines或者Columns
  private sparkShape!: Lines | Columns;

  private static defaultOptions = {
    style: {
      type: 'line',
      width: 200,
      height: 20,
      // data: [],
      isStack: false,
      color: ['#83daad', '#edbf45', '#d2cef9', '#e290b3', '#6f63f4'],
      smooth: true,
      lineStyle: {
        lineWidth: 1,
      },
      isGroup: false,
      columnStyle: {
        lineWidth: 1,
        stroke: '#fff',
        padding: 0.1,
      },
    },
  };

  constructor(options: SparklineOptions) {
    super(deepMix({}, Sparkline.defaultOptions, options));

    this.containerShape = new Rect({
      name: 'container',
      style: this.getContainerCfg(),
    });
    this.appendChild(this.containerShape);
    this.init();
  }

  public init() {
    this.sparkShape = this.createSparkline();
    this.appendChild(this.sparkShape);
  }

  /**
   * 组件的更新
   */
  public update(cfg: Partial<SparklineCfg>) {
    const { type: oldType } = this.attributes;
    this.attr(deepMix({}, this.attributes, cfg));
    const { type } = cfg;
    // 如果type变了，需要清空this.sparkShapes子元素
    if (type && type !== oldType) {
      this.clear();
      this.sparkShape = this.createSparkline();
      this.appendChild(this.sparkShape);
    }
    if (this.sparkShape instanceof Lines) this.sparkShape.update(this.getLinesCfg());
    else this.sparkShape.update(this.getColumnsCfg());
  }

  /**
   * 组件的清除
   */
  public clear() {
    this.removeChild(this.sparkShape);
    this.sparkShape.clear();
    this.sparkShape.destroy();
  }

  private createSparkline() {
    const { type } = this.attributes;
    const baseCfg = { name: 'sparkline' };
    if (type === 'line')
      return new Lines({
        ...baseCfg,
        style: this.getLinesCfg(),
      });
    // if (type === 'column')
    return new Columns({
      ...baseCfg,
      style: this.getColumnsCfg(),
    });
  }

  /**
   * 根据数据索引获取color
   */
  private getColor(index: number) {
    const { color } = this.attributes;
    if (isArray(color)) {
      return color[index % color.length];
    }
    if (isFunction(color)) {
      return color.call(null, index);
    }
    return color;
  }

  /**
   * 根据数据生成scale
   */
  private createScales(data: number[][]) {
    const { type, isGroup, barPadding } = this.attributes;
    const { width, height } = this.getContainerCfg();
    const [minVal, maxVal] = getRange(data);
    return {
      type,
      x:
        type === 'line'
          ? new Linear({
              domain: [0, data[0].length - 1],
              range: [0, width],
            })
          : new Band({
              domain: data[0].map((val, idx) => idx),
              range: [0, width],
              paddingInner: isGroup ? barPadding : 0,
            }),
      y: new Linear({
        domain: [minVal >= 0 ? 0 : minVal, maxVal],
        range: [height, 0],
      }),
    };
  }

  /**
   * 将data统一格式化为数组形式
   * 如果堆叠，则生成堆叠数据
   */
  private getData(): Data {
    const { data: _ } = this.attributes;
    if (!_ || _?.length === 0) return [[]];
    let data = clone(_);
    // number[] -> number[][]
    if (isNumber(data[0])) data = [data];
    return data;
  }

  private getContainerCfg() {
    const { width, height } = this.attributes;
    return { width, height } as { width: number; height: number };
  }

  private getLinesCfg(): ILinesCfg {
    const { areaStyle, isStack, lineStyle, smooth } = this.attributes;
    const { width } = this.getContainerCfg();
    let data = this.getData();
    if (data[0].length === 0) return { lines: [], areas: [] };
    if (isStack) data = getStackedData(data);
    const { x, y } = this.createScales(data) as { x: Linear; y: Linear };
    // 线条Path
    const lines = dataToLines(data, { type: 'line', x, y });

    // 生成区域path
    let areas: PathCommand[][] = [];
    if (areaStyle) {
      const range = getRange(data);
      const baseline = y.map(range[0] < 0 ? 0 : range[0]);
      if (isStack) {
        areas = smooth
          ? linesToStackCurveAreaPaths(lines, width, baseline)
          : linesToStackAreaPaths(lines, width, baseline);
      } else {
        areas = linesToAreaPaths(lines, smooth!, width, baseline);
      }
    }
    return {
      lines: lines.map((line, idx) => {
        return {
          stroke: this.getColor(idx),
          path: smooth ? lineToCurvePath(line) : lineToLinePath(line),
          ...lineStyle,
        };
      }),
      areas: areas.map((path, idx) => {
        return {
          path,
          fill: this.getColor(idx),
          ...areaStyle,
        };
      }),
    };
  }

  private getColumnsCfg(): IColumnsCfg {
    const { isStack, columnStyle } = this.attributes;
    const { height } = this.getContainerCfg();
    let data = this.getData();
    if (!data) return { columns: [] };
    if (isStack) data = getStackedData(data);
    const { x, y } = this.createScales(data) as { x: Band; y: Linear };
    const [minVal, maxVal] = getRange(data);
    const heightScale = new Linear({
      domain: [0, maxVal - (minVal > 0 ? 0 : minVal)],
      range: [0, height],
    });

    const bandWidth = x.getBandWidth();
    const rawData = this.getData();
    return {
      columns: data.map((column, i) => {
        return column.map((val, j) => {
          const barWidth = bandWidth / data.length;
          return {
            fill: this.getColor(i),
            ...columnStyle,
            ...(isStack
              ? {
                  x: x.map(j),
                  y: y.map(val),
                  width: bandWidth,
                  height: heightScale.map(rawData[i][j]),
                }
              : {
                  x: x.map(j) + barWidth * i,
                  y: val >= 0 ? y.map(val) : y.map(0),
                  width: barWidth,
                  height: heightScale.map(Math.abs(val)),
                }),
          } as IColumnCfg;
        });
      }),
    };
  }
}
